! Fortran 90 and PVM
! measure of communication performance
! Ping sends a message to Pong, which sends it back to Ping.
! The measured overall time is then halved.
!
! Laurent Gasser (lga@sma.ch)
! Swiss Meteorological Institute
! November 11, 1996

program Pong

implicit none

  integer :: my_Task_ID     = -1
  integer :: parent_Task_ID = -1
  integer :: largest_size   = -1
  integer, pointer :: items(:)

! body Pong
  call Get_Ready()
  call Bounce_Integers()
  call Bounce_Reals()
  call Bounce_Complexes()
  call Bounce_Logicals()
  call Bounce_Characters()
  call Take_Leave()

contains

  subroutine Get_Ready()
    use PVM_Constants
    use PVM_Interfaces
    use Politeness
    use Protocol
  ! local declarations
    integer :: buffer_ID = -1
    integer :: i         = -666
    integer :: info      = PvmOk

  ! body Get_Ready
    call pvm_mytid(my_Task_ID)
    call pvm_parent(parent_Task_ID)

    call say_hello(to = parent_Task_ID)

    call await_specifications(parent_Task_ID, items)

    ! find the largest package 
    largest_size = PROLOGUE_PACKET_SIZE
    do i = 1, size(items, 1)
       if (largest_size < items(i)) then
          largest_size = items(i)
       end if
    end do
  end subroutine Get_Ready

  subroutine Bounce_Integers()
    use Protocol
    use metrics
  ! local declarations
    integer         :: i           = -666
    integer         :: data(largest_size)
    type(durations) :: arrival   = durations(-666, -666)
    type(durations) :: departure = durations(-666, -666)
    type(durations) :: delays    = durations(-666, -666)

  ! body Bounce_Integers
    ! avoid start-up effects
    call catch_packet(from    = parent_Task_ID, &
                      packet  = data(1:PROLOGUE_PACKET_SIZE), &
                      instant = arrival)
    call throw_packet(to      = parent_Task_ID, &
                      packet  = data(1:PROLOGUE_PACKET_SIZE), &
                      instant = departure)
    do i = 1, size(items, 1)
       call catch_packet(from    = parent_Task_ID, &
                         packet  = data(1:items(i)), &
                         instant = arrival)
       call throw_packet(to      = parent_Task_ID, &
                         packet  = data(1:items(i)), &
                         instant = departure)
       ! send reflection delay
       delays%early = departure%late - arrival%early
       delays%late  = departure%early - arrival%late
       call send_bouncing_delays(parent_Task_ID, delays)
    end do
  end subroutine Bounce_Integers

  subroutine Bounce_Reals()
    use Protocol
    use metrics
  ! local declarations
    integer         :: i           = -666
    real            :: data(largest_size)
    type(durations) :: arrival   = durations(-666, -666)
    type(durations) :: departure = durations(-666, -666)
    type(durations) :: delays    = durations(-666, -666)

  ! body Bounce_Reals
    ! avoid start-up effects
    call catch_packet(from    = parent_Task_ID, &
                      packet  = data(1:PROLOGUE_PACKET_SIZE), &
                      instant = arrival)
    call throw_packet(to      = parent_Task_ID, &
                      packet  = data(1:PROLOGUE_PACKET_SIZE), &
                      instant = departure)
    do i = 1, size(items, 1)
       call catch_packet(from    = parent_Task_ID, &
                         packet  = data(1:items(i)), &
                         instant = arrival)
       call throw_packet(to      = parent_Task_ID, &
                         packet  = data(1:items(i)), &
                         instant = departure)
       ! send reflection delay
       delays%early = departure%late - arrival%early
       delays%late  = departure%early - arrival%late
       call send_bouncing_delays(parent_Task_ID, delays)
    end do
  end subroutine Bounce_Reals

  subroutine Bounce_Complexes()
    use Protocol
    use metrics
  ! local declarations
    integer         :: i           = -666
    complex         :: data(largest_size)
    type(durations) :: arrival   = durations(-666, -666)
    type(durations) :: departure = durations(-666, -666)
    type(durations) :: delays    = durations(-666, -666)

  ! body Bounce_Complexes
    ! avoid start-up effects
    call catch_packet(from    = parent_Task_ID, &
                      packet  = data(1:PROLOGUE_PACKET_SIZE), &
                      instant = arrival)
    call throw_packet(to      = parent_Task_ID, &
                      packet  = data(1:PROLOGUE_PACKET_SIZE), &
                      instant = departure)
    do i = 1, size(items, 1)
       call catch_packet(from    = parent_Task_ID, &
                         packet  = data(1:items(i)), &
                         instant = arrival)
       call throw_packet(to      = parent_Task_ID, &
                         packet  = data(1:items(i)), &
                         instant = departure)
       ! send reflection delay
       delays%early = departure%late - arrival%early
       delays%late  = departure%early - arrival%late
       call send_bouncing_delays(parent_Task_ID, delays)
    end do
  end subroutine Bounce_Complexes

  subroutine Bounce_Logicals()
    use Protocol
    use metrics
  ! local declarations
    integer         :: i           = -666
    logical         :: data(largest_size)
    type(durations) :: arrival   = durations(-666, -666)
    type(durations) :: departure = durations(-666, -666)
    type(durations) :: delays    = durations(-666, -666)

  ! body Bounce_Logicals
    ! avoid start-up effects
    call catch_packet(from    = parent_Task_ID, &
                      packet  = data(1:PROLOGUE_PACKET_SIZE), &
                      instant = arrival)
    call throw_packet(to      = parent_Task_ID, &
                      packet  = data(1:PROLOGUE_PACKET_SIZE), &
                      instant = departure)
    do i = 1, size(items, 1)
       call catch_packet(from    = parent_Task_ID, &
                         packet  = data(1:items(i)), &
                         instant = arrival)
       call throw_packet(to      = parent_Task_ID, &
                         packet  = data(1:items(i)), &
                         instant = departure)
       ! send reflection delay
       delays%early = departure%late - arrival%early
       delays%late  = departure%early - arrival%late
       call send_bouncing_delays(parent_Task_ID, delays)
    end do
  end subroutine Bounce_Logicals

  subroutine Bounce_Characters()
    use Protocol
    use metrics
  ! local declarations
    integer         :: i           = -666
    character       :: data(4*largest_size)
    type(durations) :: arrival   = durations(-666, -666)
    type(durations) :: departure = durations(-666, -666)
    type(durations) :: delays    = durations(-666, -666)

  ! body Bounce_Characters
    ! avoid start-up effects
    call catch_packet(from    = parent_Task_ID, &
                      packet  = data(1:PROLOGUE_PACKET_SIZE), &
                      instant = arrival)
    call throw_packet(to      = parent_Task_ID, &
                      packet  = data(1:PROLOGUE_PACKET_SIZE), &
                      instant = departure)
    do i = 1, size(items, 1)
       call catch_packet(from    = parent_Task_ID, &
                         packet  = data(1:items(i)), &
                         instant = arrival)
       call throw_packet(to      = parent_Task_ID, &
                         packet  = data(1:items(i)), &
                         instant = departure)
       ! send reflection delay
       delays%early = departure%late - arrival%early
       delays%late  = departure%early - arrival%late
       call send_bouncing_delays(parent_Task_ID, delays)
    end do
  end subroutine Bounce_Characters

  subroutine Take_Leave()
    use PVM_Constants
    use Politeness
  ! body Take_Leave
    call say_goodbye(to = parent_Task_ID)
  end subroutine Take_Leave

end program Pong
